@@ -20,11 +20,17 @@ from __future__ import division, print_function, absolute_import
 import sys
 from enum import IntEnum
 
+import attr
+
 from hypothesis.errors import Frozen, StopTest, InvalidArgument
 from hypothesis.internal.compat import hbytes, hrange, text_type, \
     bit_length, benchmark_time, int_from_bytes, unicode_safe_repr
 from hypothesis.internal.coverage import IN_COVERAGE_TESTS
 from hypothesis.internal.escalation import mark_for_escalation
+from hypothesis.internal.conjecture.utils import calc_label_from_name
+
+TOP_LABEL = calc_label_from_name('top')
+DRAW_BYTES_LABEL = calc_label_from_name('draw_bytes() in ConjectureData')
 
 
 class Status(IntEnum):
@@ -33,6 +39,46 @@ class Status(IntEnum):
     VALID = 2
     INTERESTING = 3
 
+    def __repr__(self):
+        return 'Status.%s' % (self.name,)
+
+
+@attr.s(slots=True)
+class Example(object):
+    depth = attr.ib()
+    label = attr.ib()
+    index = attr.ib()
+    start = attr.ib()
+    end = attr.ib(default=None)
+
+    # An example is "trivial" if it only contains forced bytes and zero bytes.
+    # All examples start out as trivial, and then get marked non-trivial when
+    # we see a byte that is neither forced nor zero.
+    trivial = attr.ib(default=True)
+    discarded = attr.ib(default=None)
+    children = attr.ib(default=attr.Factory(list))
+
+    @property
+    def length(self):
+        return self.end - self.start
+
+
+@attr.s(hash=False, cmp=False, slots=True)
+class StructuralTag(object):
+    label = attr.ib()
+
+
+STRUCTURAL_TAGS = {}  # type: dict
+
+
+def structural_tag(label):
+    try:
+        return STRUCTURAL_TAGS[label]
+    except KeyError:
+        result = StructuralTag(label)
+        STRUCTURAL_TAGS[label] = result
+        return result
+
 
 global_test_counter = 0
 
@@ -56,28 +102,30 @@ class ConjectureData(object):
         self.is_find = False
         self._draw_bytes = draw_bytes
         self.overdraw = 0
-        self.level = 0
         self.block_starts = {}
         self.blocks = []
         self.buffer = bytearray()
         self.output = u''
         self.status = Status.VALID
         self.frozen = False
-        self.intervals_by_level = []
-        self.interval_stack = []
         global global_test_counter
         self.testcounter = global_test_counter
         global_test_counter += 1
         self.start_time = benchmark_time()
         self.events = set()
         self.forced_indices = set()
-        self.capped_indices = {}
+        self.forced_blocks = set()
+        self.masked_indices = {}
         self.interesting_origin = None
         self.tags = set()
         self.draw_times = []
-        self.__intervals = None
-        self.shrinking_blocks = set()
-        self.discarded = []
+        self.max_depth = 0
+
+        self.examples = []
+        self.example_stack = []
+        self.has_discards = False
+
+        self.start_example(TOP_LABEL)
 
     def __assert_not_frozen(self, name):
         if self.frozen:
@@ -90,7 +138,9 @@ class ConjectureData(object):
 
     @property
     def depth(self):
-        return len(self.interval_stack)
+        # We always have a single example wrapping everything. We want to treat
+        # that as depth 0 rather than depth 1.
+        return len(self.example_stack) - 1
 
     @property
     def index(self):
@@ -102,7 +152,7 @@ class ConjectureData(object):
             value = unicode_safe_repr(value)
         self.output += value
 
-    def draw(self, strategy):
+    def draw(self, strategy, label=None):
         if self.is_find and not strategy.supports_find:
             raise InvalidArgument((
                 'Cannot use strategy %r within a call to find (presumably '
@@ -119,15 +169,17 @@ class ConjectureData(object):
             original_tracer = sys.gettrace()
             try:
                 sys.settrace(None)
-                return self.__draw(strategy)
+                return self.__draw(strategy, label=label)
             finally:
                 sys.settrace(original_tracer)
         else:
-            return self.__draw(strategy)
+            return self.__draw(strategy, label=label)
 
-    def __draw(self, strategy):
+    def __draw(self, strategy, label):
         at_top_level = self.depth == 0
-        self.start_example()
+        if label is None:
+            label = strategy.label
+        self.start_example(label=label)
         try:
             if not at_top_level:
                 return strategy.do_draw(self)
@@ -141,69 +193,77 @@ class ConjectureData(object):
                 finally:
                     self.draw_times.append(benchmark_time() - start_time)
         finally:
-            if not self.frozen:
-                self.stop_example()
+            self.stop_example()
 
-    def start_example(self):
+    def start_example(self, label):
         self.__assert_not_frozen('start_example')
-        self.interval_stack.append(self.index)
-        self.level += 1
+
+        i = len(self.examples)
+        new_depth = self.depth + 1
+        ex = Example(
+            index=i,
+            depth=new_depth, label=label, start=self.index,
+        )
+        self.examples.append(ex)
+        if self.example_stack:
+            p = self.example_stack[-1]
+            self.examples[p].children.append(ex)
+        self.example_stack.append(i)
+        self.max_depth = max(self.max_depth, self.depth)
+        return ex
 
     def stop_example(self, discard=False):
         if self.frozen:
             return
-        self.level -= 1
-        while self.level >= len(self.intervals_by_level):
-            self.intervals_by_level.append([])
-        k = self.interval_stack.pop()
-        if k != self.index:
-            t = (k, self.index)
-            self.intervals_by_level[self.level].append(t)
-            if discard:
-                while self.discarded:
-                    u, v = t
-                    r, s = self.discarded[-1]
-                    if u <= r <= s <= v:
-                        self.discarded.pop()
-                    else:
-                        assert s <= u
-                        break
-                self.discarded.append(t)
+
+        k = self.example_stack.pop()
+        ex = self.examples[k]
+        ex.end = self.index
+
+        if self.example_stack and not ex.trivial:
+            self.examples[self.example_stack[-1]].trivial = False
+
+        # We don't want to count empty examples as discards even if the flag
+        # says we should. This leads to situations like
+        # https://github.com/HypothesisWorks/hypothesis/issues/1230
+        # where it can look like we should discard data but there's nothing
+        # useful for us to do.
+        if self.index == ex.start:
+            discard = False
+
+        ex.discarded = discard
+
+        if discard:
+            self.has_discards = True
 
     def note_event(self, event):
         self.events.add(event)
 
-    @property
-    def intervals(self):
-        assert self.frozen
-        if self.__intervals is None:
-            intervals = set(self.blocks)
-            for l in self.intervals_by_level:
-                intervals.update(l)
-                for i in hrange(len(l) - 1):
-                    if (
-                        l[i] not in self.discarded and
-                        l[i + 1] not in self.discarded and
-                        l[i][1] == l[i + 1][0]
-                    ):
-                        intervals.add((l[i][0], l[i + 1][1]))
-            for i in hrange(len(self.blocks) - 1):
-                intervals.add((self.blocks[i][0], self.blocks[i + 1][1]))
-            # Intervals are sorted as longest first, then by interval start.
-            self.__intervals = tuple(sorted(
-                set(intervals),
-                key=lambda se: (se[0] - se[1], se[0])
-            ))
-            del self.intervals_by_level
-        return self.__intervals
-
     def freeze(self):
         if self.frozen:
             assert isinstance(self.buffer, hbytes)
             return
-        self.frozen = True
         self.finish_time = benchmark_time()
 
+        while self.example_stack:
+            self.stop_example()
+
+        self.frozen = True
+
+        if self.status >= Status.VALID:
+            discards = []
+            for ex in self.examples:
+                if ex.length == 0:
+                    continue
+                if discards:
+                    u, v = discards[-1]
+                    if u <= ex.start <= ex.end <= v:
+                        continue
+                if ex.discarded:
+                    discards.append((ex.start, ex.end))
+                    continue
+                self.tags.add(structural_tag(ex.label))
+
         self.buffer = hbytes(self.buffer)
         self.events = frozenset(self.events)
         del self._draw_bytes
@@ -221,10 +281,11 @@ class ConjectureData(object):
             assert len(buf) == n_bytes
             mask = (1 << (n % 8)) - 1
             buf[0] &= mask
-            self.capped_indices[self.index] = mask
+            self.masked_indices[self.index] = mask
             buf = hbytes(buf)
             self.__write(buf)
             result = int_from_bytes(buf)
+
         assert bit_length(result) <= n
         return result
 
@@ -233,8 +294,9 @@ class ConjectureData(object):
         self.__check_capacity(len(string))
         assert isinstance(string, hbytes)
         original = self.index
-        self.__write(string)
+        self.__write(string, forced=True)
         self.forced_indices.update(hrange(original, self.index))
+        self.forced_blocks.add(len(self.blocks) - 1)
         return string
 
     def __check_capacity(self, n):
@@ -244,7 +306,9 @@ class ConjectureData(object):
             self.freeze()
             raise StopTest(self.testcounter)
 
-    def __write(self, result):
+    def __write(self, result, forced=False):
+        ex = self.start_example(DRAW_BYTES_LABEL)
+        ex.trivial = forced or not any(result)
         initial = self.index
         n = len(result)
         self.block_starts.setdefault(n, []).append(initial)
@@ -252,6 +316,7 @@ class ConjectureData(object):
         assert len(result) == n
         assert self.index == initial
         self.buffer.extend(result)
+        self.stop_example()
 
     def draw_bytes(self, n):
         self.__assert_not_frozen('draw_bytes')
